# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from os import access, R_OK, W_OK, X_OK
from pathlib import Path

# Geode
from geode_gem import __main__ as geode_main

# System
from tempfile import TemporaryDirectory

# Unittest
import unittest


# ------------------------------------------------------------------------------
#   Tests
# ------------------------------------------------------------------------------

class GeodeGEMMainTC(unittest.TestCase):

    def setUp(self):
        """ Initialize each test with some data
        """

        self.temp_directory = Path(TemporaryDirectory().name)
        self.temp_directory.mkdir()

    def tearDown(self):
        """ Remove data from initialization when test terminate
        """

        for filename in self.temp_directory.iterdir():
            filename.unlink()

        self.temp_directory.rmdir()

    def test_copy_file(self):
        """ Check geode_gem.__main__.copy_file method
        """

        path_mario = self.temp_directory.joinpath("mario-is-missing")
        path_luigi = self.temp_directory.joinpath("luigi-is-coming")

        self.assertFalse(geode_main.copy_file(path_mario, path_luigi))
        self.assertFalse(path_mario.exists())
        self.assertFalse(path_luigi.exists())

        path_mario.write_text("mamma mia")
        self.assertTrue(geode_main.copy_file(path_mario, path_luigi))
        self.assertTrue(path_mario.exists())
        self.assertTrue(path_luigi.exists())
        self.assertEqual(path_mario.read_text(), path_luigi.read_text())

        path_toad = self.temp_directory.joinpath("toad", "wrong-castle")
        with self.assertRaises(FileNotFoundError):
            geode_main.copy_file(path_mario, path_toad)

        path_toad.parent.mkdir()
        self.assertTrue(geode_main.copy_file(path_mario, path_toad))

        path_toad.unlink()
        path_toad.parent.rmdir()

    def test_make_directory(self):
        """ Check geode_gem.__main__.make_directory method
        """

        self.assertFalse(geode_main.make_directory(self.temp_directory))

        path = self.temp_directory.joinpath("geode-gem-pytest-directory")
        self.assertTrue(geode_main.make_directory(path))
        self.assertTrue(path.exists())
        self.assertTrue(path.is_dir())
        self.assertTrue(access(path, R_OK))
        self.assertTrue(access(path, W_OK))
        self.assertTrue(access(path, X_OK))
        path.rmdir()

        subpath = path.joinpath("subdirectory")
        with self.assertRaises(FileNotFoundError):
            geode_main.make_directory(subpath, make_parents=False)

        self.assertTrue(geode_main.make_directory(subpath, make_parents=True))
        self.assertTrue(subpath.exists())
        subpath.rmdir()
        path.rmdir()

        self.assertTrue(geode_main.make_directory(path, mode=0o600))
        self.assertTrue(path.exists())
        self.assertTrue(access(path, R_OK))
        self.assertTrue(access(path, W_OK))
        self.assertFalse(access(path, X_OK))
        path.rmdir()

    def test_init_environment(self):
        """ Check geode_gem.__main__.init_environment method
        """

        metadata = geode_main.init_environment()

        from geode_gem.ui.data import Columns, Icons, Metadata

        data = [
            (Columns.List, "list", str()),
            (Columns.Grid, "grid", str()),
            (Icons, "icons", str()),
            (Icons.Symbolic, "icons", "-symbolic"),
            (Icons.Size, "icon-sizes", str()),
            (Metadata, "metadata", str()),
        ]

        for (object_type, section, suffix) in data:
            self.assertIn(section, metadata.sections())

            for key, value in object_type.__dict__.items():
                if not isinstance(value, str) or key.startswith("__"):
                    continue

                self.assertIn(key.lower(), metadata.options(section))
                self.assertEqual(metadata.get(section, key.lower()),
                                 value.replace(suffix, str()))

        self.assertEqual(Columns.ORDER, metadata.get("misc", "columns_order"))
