# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Datetime
import datetime

# Filesystem
from pathlib import Path

# Geode
from geode_gem.engine.game import Game
from geode_gem.ui import utils

# GObject
from gi.repository import GdkPixbuf, Gtk

# System
from tempfile import TemporaryDirectory

# Unittest
import unittest
from unittest import mock


# ------------------------------------------------------------------------------
#   Tests
# ------------------------------------------------------------------------------

class GeodeGEMUIUtilsTC(unittest.TestCase):

    def setUp(self):
        """ Initialize each test with some data
        """

        self.temp_directory = Path(TemporaryDirectory().name)
        self.temp_directory.mkdir()

        self.prefer_dark_theme = "gtk-application-prefer-dark-theme"

    def tearDown(self):
        """ Remove data from initialization when test terminate
        """

        for filename in self.temp_directory.iterdir():
            filename.unlink()

        self.temp_directory.rmdir()

    def test_call_external_application(self):
        """ Check geode_gem.ui.utils.call_external_application method
        """

        output = utils.call_external_application("echo", "hello world")
        self.assertIsNotNone(output)
        self.assertEqual(output, "hello world")

        output = utils.call_external_application("ls", "-help")
        self.assertIsNotNone(output)
        self.assertTrue(output.endswith("-- 'e'"))

    def test_check_game_is_editable(self):
        """ Check geode_gem.ui.utils.check_game_is_editable method
        """

        path = self.temp_directory.joinpath("game_file.txt")
        path.write_text("All your bases are belong to us!")

        self.assertTrue(utils.check_game_is_editable(Game(None, path)))

        path.chmod(0o600)
        self.assertTrue(utils.check_game_is_editable(Game(None, path)))

        path.chmod(0o400)
        self.assertFalse(utils.check_game_is_editable(Game(None, path)))

    def test_check_mednafen(self):
        """ Check geode_gem.ui.utils.check_mednafen method
        """

        with mock.patch("geode_gem.ui.utils.call_external_application",
                        return_value="Wrong output\n"):
            self.assertFalse(utils.check_mednafen())

        with mock.patch("geode_gem.ui.utils.call_external_application",
                        return_value="Starting Mednafen 1.0.0\n"):
            self.assertTrue(utils.check_mednafen())

    def test_generate_blank_icons(self):
        """ Check geode_gem.ui.utils.generate_blank_icons method
        """

        icons = utils.generate_blank_icons()
        self.assertEqual(len(icons), 8)

        for icon_size in icons.keys():
            icon = icons.get(icon_size)

            self.assertIsNotNone(icon)
            self.assertEqual(icon.get_bits_per_sample(), 8)
            self.assertEqual(icon.get_colorspace(), GdkPixbuf.Colorspace.RGB)
            self.assertTrue(icon.get_has_alpha())
            self.assertEqual(icon.get_height(), icon_size)
            self.assertEqual(icon.get_width(), icon_size)

    def test_icon_from_data(self):
        """ Check geode_gem.ui.utils.icon_from_data method
        """

        fallback = utils.icon_from_data(Path("unkwown-icon-filename.png"))
        self.assertIsInstance(fallback, GdkPixbuf.Pixbuf)
        self.assertEqual(fallback.get_bits_per_sample(), 8)
        self.assertEqual(fallback.get_colorspace(), GdkPixbuf.Colorspace.RGB)
        self.assertTrue(fallback.get_has_alpha())
        self.assertEqual(fallback.get_height(), 24)
        self.assertEqual(fallback.get_width(), 24)

        icon = utils.icon_from_data(Path("another-one-unkwown.png"),
                                    fallback=fallback, width=32, height=32)
        self.assertIsInstance(icon, GdkPixbuf.Pixbuf)
        self.assertEqual(icon.get_height(), 24)
        self.assertEqual(icon.get_width(), 24)

        icon = utils.icon_from_data(Path("test", "data", "geode_gem_logo.png"),
                                    fallback=fallback, width=32, height=32)
        self.assertIsInstance(icon, GdkPixbuf.Pixbuf)
        self.assertEqual(icon.get_height(), 32)
        self.assertEqual(icon.get_width(), 32)

    def test_magic_from_file(self):
        """ Check geode_gem.ui.utils.magic_from_file method

        Needs 'file' application to check correctly output results
        """

        filetext_empty = self.temp_directory.joinpath("filetext_empty.txt")
        filetext_empty.touch()

        filetext_1 = self.temp_directory.joinpath("filetext1.txt")
        filetext_1.write_bytes(b"This is a file without line terminator!")

        filetext_2 = self.temp_directory.joinpath("filetext2.txt")
        filetext_2.write_bytes(b"This is a file with a line terminator!\n")

        data = [
            (filetext_empty, "inode/x-empty", "empty"),
            (filetext_1, "text/plain", "ASCII text, with no line terminators"),
            (filetext_2, "text/plain", "ASCII text"),
            (Path("test", "data", "geode_gem_logo.png"), "image/png",
             "PNG image data, 32 x 32, 8-bit/color RGBA, non-interlaced"),
            (Path("test", "test_ui_utils.py"), "text/x-script.python",
             "Python script, ASCII text executable"),
        ]

        for filename, mime_type, human_text in data:
            self.assertTrue(filename.exists())

            for result in (mime_type, human_text):
                self.assertEqual(utils.magic_from_file(
                    filename, mime=(result == mime_type)), result)

    def test_on_prefer_dark_theme(self):
        """ Check geode_gem.ui.utils.on_prefer_dark_theme method
        """

        utils.on_prefer_dark_theme(True)
        self.assertTrue(
            Gtk.Settings.get_default().get_property(self.prefer_dark_theme))

        utils.on_prefer_dark_theme(False)
        self.assertFalse(
            Gtk.Settings.get_default().get_property(self.prefer_dark_theme))

    def test_on_entry_clear(self):
        """ Check geode_gem.ui.utils.on_entry_clear method
        """

        self.assertFalse(utils.on_entry_clear(Gtk.Label(), None, None))

        entry_widget = Gtk.Entry()
        self.assertFalse(utils.on_entry_clear(
            entry_widget, Gtk.EntryIconPosition.PRIMARY, None))
        self.assertFalse(utils.on_entry_clear(
            entry_widget, Gtk.EntryIconPosition.SECONDARY, None))

        entry_widget.set_text("The Antidote")
        self.assertFalse(utils.on_entry_clear(
            entry_widget, Gtk.EntryIconPosition.PRIMARY, None))
        self.assertTrue(utils.on_entry_clear(
            entry_widget, Gtk.EntryIconPosition.SECONDARY, None))

    def test_replace_for_markup(self):
        """ Check geode_gem.ui.utils.replace_for_markup method
        """

        data = [
            ('&', "&amp;"),
            ('<', "&lt;"),
            ('>', "&gt;"),
            ("</html>", "&lt;/html&gt;"),
            ("Tix & Romaric", "Tix &amp; Romaric"),
        ]

        for text, result in data:
            self.assertEqual(utils.replace_for_markup(text), result)

    @mock.patch("geode_gem.ui.utils.get_today_from_datetime")
    def test_string_from_date(self, mock_object):
        """ Check geode_gem.ui.utils.string_from_date method

        A patch is mandatory to test string_from_date function, since a
        datetime.now() is called
        """

        mock_object.return_value = datetime.datetime(2020, 6, 30)

        self.assertIsNone(utils.string_from_date(None))

        data = [
            (datetime.datetime(2018, 1, 1), "2 years ago"),
            (datetime.datetime(2019, 1, 1), "Last year"),
            (datetime.datetime(2020, 1, 1), "6 months ago"),
            (datetime.datetime(2020, 5, 29), "Last month"),
            (datetime.datetime(2020, 6, 20), "10 days ago"),
            (datetime.datetime(2020, 6, 29), "Yesterday"),
            (datetime.datetime(2020, 6, 30), "Today"),
            (datetime.datetime(2020, 7, 1), "Tomorrow"),
            (datetime.datetime(2020, 7, 15), "In 15 days"),
            (datetime.datetime(2020, 7, 30), "Next month"),
            (datetime.datetime(2020, 8, 30), "In 2 months"),
            (datetime.datetime(2021, 8, 30), "Next year"),
            (datetime.datetime(2022, 8, 30), "In 2 years"),
        ]

        for delta, output in data:
            self.assertEqual(utils.string_from_date(delta), output)

    def test_string_from_time(self):
        """ Check geode_gem.ui.utils.string_from_time method
        """

        data = [
            (datetime.timedelta(seconds=1), "1 second"),
            (datetime.timedelta(seconds=42), "42 seconds"),
            (datetime.timedelta(seconds=61), "1 minute"),
            (datetime.timedelta(minutes=42), "42 minutes"),
            (datetime.timedelta(hours=1), "1 hour"),
            (datetime.timedelta(hours=42), "42 hours"),
            (datetime.timedelta(days=42), "1008 hours"),
            (datetime.timedelta(days=42, hours=24), "1032 hours"),
        ]

        for delta, output in data:
            self.assertEqual(utils.string_from_time(delta), output)

    def test_get_filename_from_icon_theme(self):
        """ Check geode_gem.ui.utils.get_filename_from_icon_theme method
        """

        data = ("folder", "terminal", "go-up", "document-new")

        icons_folder = self.temp_directory.joinpath("22x22", "apps")
        icons_folder.mkdir(parents=True)

        for icon_name in data:
            icons_folder.joinpath(f"{icon_name}.png").touch()

        # Use custom folder in temporary storage as default icons search path
        icon_theme = Gtk.IconTheme.get_default()
        icon_theme.set_search_path([str(icons_folder)])

        for icon_name in data:
            path = utils.get_filename_from_icon_theme(icon_theme, icon_name, 22)
            self.assertIsNotNone(path)
            self.assertTrue(Path(path).exists())
            self.assertEqual(
                path, str(icons_folder.joinpath(f"{icon_name}.png")))

        # Remove generated icons
        for filename in icons_folder.iterdir():
            filename.unlink()

        icons_folder.rmdir()
        icons_folder.parent.rmdir()


class GeodeGEMUIUtilsPixbufSizeTC(unittest.TestCase):

    def setUp(self):
        """ Initialize each test with some data
        """

        self.pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(
            str(Path("test", "data", "geode_gem_logo.png")), 96, 96)

    def test_constructor(self):
        """ Check geode_gem.ui.utils.PixbufSize constructor
        """

        pixbuf_size = utils.PixbufSize(self.pixbuf, 800)

        self.assertEqual(pixbuf_size.width, 96)
        self.assertEqual(pixbuf_size.height, 96)
        self.assertEqual(pixbuf_size.size, (96, 96))

    def test_default_width(self):
        """ Check geode_gem.ui.utils.PixbufSize.default_width method
        """

        pixbuf_size = utils.PixbufSize(self.pixbuf, 800)
        self.assertEqual(pixbuf_size.default_width, 800)

        pixbuf_size.width = 80
        self.assertEqual(pixbuf_size.default_width, 666)

        pixbuf_size.width = 96
        pixbuf_size.height = 80
        self.assertEqual(pixbuf_size.default_width, 800)

    def test_default_height(self):
        """ Check geode_gem.ui.utils.PixbufSize.default_height method
        """

        pixbuf_size = utils.PixbufSize(self.pixbuf, 800)
        self.assertEqual(pixbuf_size.default_height, 800)

        pixbuf_size.width = 80
        self.assertEqual(pixbuf_size.default_height, 800)

        pixbuf_size.width = 96
        pixbuf_size.height = 80
        self.assertEqual(pixbuf_size.default_height, 666)

    def test_get_ratio(self):
        """ Check geode_gem.ui.utils.PixbufSize.get_ratio method
        """

        pixbuf_size = utils.PixbufSize(self.pixbuf, 800)

        ratio_w, ratio_h = pixbuf_size.get_ratio(1024, 768)
        self.assertAlmostEqual(ratio_w, 1.28)
        self.assertAlmostEqual(ratio_h, 0.96)

        pixbuf_size.width = 80
        ratio_w, ratio_h = pixbuf_size.get_ratio(1024, 768)
        self.assertAlmostEqual(ratio_w, 1.54, places=2)
        self.assertAlmostEqual(ratio_h, 0.96)

        pixbuf_size.width = 256
        pixbuf_size.height = 72
        ratio_w, ratio_h = pixbuf_size.get_ratio(1024, 768)
        self.assertAlmostEqual(ratio_w, 1.28)
        self.assertAlmostEqual(ratio_h, 3.41, places=2)
