# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon

# GObject
from gi.repository import Gtk, Pango

# Unittest
import unittest


# ------------------------------------------------------------------------------
#   Tests
# ------------------------------------------------------------------------------

class DummyGeodeGtkBox(GeodeGtkCommon, Gtk.Box):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Box, **kwargs)

        for element in args:
            self.append_widget(element)
            self.add(element)


class DummyGeodeGtkLabel(GeodeGtkCommon, Gtk.Label):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Label, **kwargs)


class DummyGeodeGtkLabelWithSetters(GeodeGtkCommon, Gtk.Label):

    __setters__ = {
        "set_use_markup": True,
        "set_ellipsize": Pango.EllipsizeMode.END,
    }

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Label, **kwargs)


class GeodeGEMWidgetsCommonTC(unittest.TestCase):

    def check_visible_status(self, widget):
        """ Toggle visible status to check widget visibility status

        Parameters
        ----------
        widget: Gtk.Widget
            Object which receive signal
        """

        self.assertFalse(widget.get_visible())
        widget.show()
        self.assertTrue(widget.get_visible())

    def test_base_class_constructor(self):
        """ Check geode_gem.widgets.common.GeodeGtkCommon.__init__ methods
        """

        with self.assertRaises(TypeError):
            GeodeGtkCommon()

        with self.assertRaises(ValueError):
            GeodeGtkCommon(None)

        with self.assertRaises(TypeError):
            GeodeGtkCommon(Gtk.Label)

    def test_dummy_class(self):
        """ Check geode_gem.widgets.common.GeodeGtkCommon.__init__ methods
        """

        widget = DummyGeodeGtkLabel()
        self.assertFalse(widget.get_use_markup())
        self.assertEqual(widget.get_ellipsize(), Pango.EllipsizeMode.NONE)
        self.check_visible_status(widget)

        widget = DummyGeodeGtkBox()
        self.assertEqual(widget.get_spacing(), 0)
        self.assertFalse(widget.get_homogeneous())
        self.check_visible_status(widget)

        with self.assertRaises(KeyError):
            widget.get_widget("unknown-widget")

    def test_dummy_class_with_setters(self):
        """ Check geode_gem.widgets.common.GeodeGtkCommon.__init__ methods
        """

        widget = DummyGeodeGtkLabelWithSetters()
        self.assertTrue(widget.get_use_markup())
        self.assertEqual(widget.get_ellipsize(), Pango.EllipsizeMode.END)
        self.check_visible_status(widget)

    def test_dummy_class_with_kwargs(self):
        """ Check geode_gem.widgets.common.GeodeGtkCommon.__init__ methods
        """

        widget = DummyGeodeGtkLabel(
            set_text="label", set_style=Gtk.STYLE_CLASS_SUGGESTED_ACTION)
        self.assertEqual(widget.get_text(), "label")
        self.assertEqual(
            widget.current_style, Gtk.STYLE_CLASS_SUGGESTED_ACTION)

    def test_hierarchical_dummy_class(self):
        """ Check geode_gem.widgets.common.GeodeGtkCommon widget methods
        """

        box = DummyGeodeGtkBox(DummyGeodeGtkLabel(identifier="test-label"))
        self.check_visible_status(box)
        self.assertTrue(box.has_widget("test-label"))

        first_widget = box.get_widget("test-label")
        self.assertIs(type(first_widget), DummyGeodeGtkLabel)
        self.assertEqual(first_widget.identifier, "test-label")

        second_widget = DummyGeodeGtkLabel(set_text="label", identifier="42")
        box.append_widget(second_widget)
        self.assertTrue(box.has_widget("test-label"))
        self.assertTrue(box.has_widget("42"))
        self.assertListEqual(box.get_children(), [first_widget])

        box.add(second_widget)
        self.assertListEqual(
            box.get_children(), [first_widget, second_widget])

        first_widget.set_sensitive(False)
        self.assertFalse(first_widget.get_sensitive())
        self.assertFalse(box.get_sensitive(widget=first_widget.identifier))

        box.set_sensitive(True, widget=first_widget.identifier)
        self.assertTrue(first_widget.get_sensitive())

    def test_dummy_class_style_setting(self):
        """ Check geode_gem.widgets.common.GeodeGtkCommon style methods
        """

        widget = DummyGeodeGtkLabel()
        self.check_visible_status(widget)
        widget.set_style(Gtk.STYLE_CLASS_SUGGESTED_ACTION)

        self.assertIn(Gtk.STYLE_CLASS_SUGGESTED_ACTION, widget.get_style())

        widget.identifier = "test-label"
        box = DummyGeodeGtkBox(widget)
        self.check_visible_status(box)

        first_widget = box.get_widget("test-label")
        self.assertEqual(widget, first_widget)

        first_widget.set_style()
        self.assertFalse(widget.has_style(Gtk.STYLE_CLASS_SUGGESTED_ACTION))
